@@ -13,7 +13,7 @@ import random
 import re
 import shlex
 import ssl
-import subprocess
+import subprocess  # noqa: S404
 import sys
 import tempfile
 import typing
@@ -43,6 +43,38 @@ RE_VERSION: Final = re.compile(
 RE_LINE_IDX: Final = re.compile(r" ^ Hello \s+ (?P<idx> 0 | [1-9][0-9]* ) $ ", re.X)
 
 
+@dataclasses.dataclass
+class StRunError(Exception):
+    """An exception that occurred during the `stunnel` testing."""
+
+
+@dataclasses.dataclass
+class NoListeningPortError(StRunError):
+    """Could not find a port to listen on."""
+
+    def __str__(self) -> str:
+        """Provide a human-readable description of the error."""
+        return "Could not find a port to listen on"
+
+
+@dataclasses.dataclass
+class NoListeningSocketError(StRunError):
+    """The listening socket setup failed."""
+
+    def __str__(self) -> str:
+        """Provide a human-readable description of the error."""
+        return "listener] Expected a listening socket to be created already"
+
+
+@dataclasses.dataclass
+class UnexpectedChildError(StRunError):
+    """The child process terminated unexpectedly."""
+
+    def __str__(self) -> str:
+        """Provide a human-readable description of the error."""
+        return "The stunnel child process ended unexpectedly"
+
+
 @dataclasses.dataclass(frozen=True)
 class Event:
     """The base class for an event."""
@@ -222,7 +254,7 @@ async def get_stunnel_version(cfg: Config) -> str:
         match = RE_VERSION.match(line)
         if match:
             return match.group("version")
-    sys.exit("Could not find the version line in the `{cmd_str}` output:\n" + "\n".join(lines))
+    sys.exit(f"Could not find the version line in the `{cmd_str}` output:\n" + "\n".join(lines))
 
 
 async def find_listening_port(
@@ -246,7 +278,7 @@ async def find_listening_port(
         except OSError as err:
             print(f"[listener] - {port} failed: {err}")
 
-    raise Exception("Could not find a port to listen on")
+    raise NoListeningPortError
 
 
 async def start_listener(cfg: Config) -> None:
@@ -264,7 +296,7 @@ async def start_listener(cfg: Config) -> None:
                 FatalEvent(
                     etype="listener",
                     reason=f"Handling a new connection: {err}",
-                )
+                ),
             )
             return
 
@@ -285,7 +317,7 @@ async def start_listener(cfg: Config) -> None:
                     peer=peer,
                     line=line,
                     idx=int(match.group("idx")),
-                )
+                ),
             )
 
             print(f"[{peer}] Writing something back...")
@@ -307,7 +339,7 @@ async def start_listener(cfg: Config) -> None:
     try:
         srv: Final = await find_listening_port(6502, 6502 + 1000, client_connected)
         if not srv.sockets:
-            raise Exception(f"[listener] Expected a listening socket, got {srv.sockets!r}")
+            raise NoListeningSocketError  # noqa: TRY301
         hostname, port = srv.sockets[0].getsockname()[:2]
         print(f"[listener] Telling the main thread about [{hostname}]:{port}")
         await cfg.mainq.put(ListenerStartedEvent(etype="listener", hostname=hostname, port=port))
@@ -384,7 +416,7 @@ async def expect_client_event(
     return evt
 
 
-async def test_connect(cfg: Config, conn: TestConnection) -> None:  # noqa: PLR0915
+async def test_connect(cfg: Config, conn: TestConnection) -> None:  # noqa: PLR0912,PLR0915
     """Connect to the stunnel listener, read/write data as requested."""
 
     async def wait() -> None:
@@ -435,13 +467,15 @@ async def test_connect(cfg: Config, conn: TestConnection) -> None:  # noqa: PLR0
         if conn.encrypted:
             if local == evt_conn.peer:
                 sys.exit(f"[{tag}] expected something other than {local!r}")
-        else:
-            if local != evt_conn.peer:  # noqa: PLR5501  # clarity
+        else:  # noqa: PLR5501  # clarity
+            if local != evt_conn.peer:
                 sys.exit(f"[{tag}] expected {local!r}, got {evt_conn.peer!r}")
 
         print(f"[{tag}] Waiting for the server to receive it...")
         evt_recv: Final = await expect_client_event(
-            conn.msgq, "test_connect", ClientReceivedDataEvent
+            conn.msgq,
+            "test_connect",
+            ClientReceivedDataEvent,
         )
         if evt_recv.line != line:
             sys.exit(f"[{tag}] Send: expected {line!r}, got {evt_recv.line!r}")
@@ -469,7 +503,7 @@ async def test_connect(cfg: Config, conn: TestConnection) -> None:  # noqa: PLR0
         print(f"[{tag}] Done")
     except Exception as err:  # noqa: BLE001
         print(f"[{tag}] Telling the main thread about {err}")
-        await cfg.mainq.put(FatalEvent(etype=tag, reason="Something went wrong: {err}"))
+        await cfg.mainq.put(FatalEvent(etype=tag, reason=f"Something went wrong: {err}"))
 
 
 async def find_stunnel_port(cfg: Config, hostname: str, port: int) -> int:
@@ -485,7 +519,7 @@ async def find_stunnel_port(cfg: Config, hostname: str, port: int) -> int:
             FatalEvent(
                 etype="find_stunnel_port",
                 reason=f"callback invoked: reader {reader!r} writer {writer!r}",
-            )
+            ),
         )
 
     srv: Final = await find_listening_port(port + 1, port + 1001, callback, hostname=hostname)
@@ -499,7 +533,10 @@ async def find_stunnel_port(cfg: Config, hostname: str, port: int) -> int:
 
 
 async def prepare_config_file(
-    cfg: Config, hostname: str, port: int, stunnel_port: int
+    cfg: Config,
+    hostname: str,
+    port: int,
+    stunnel_port: int,
 ) -> pathlib.Path:
     """Create a configuration file for stunnel."""
     proc: Final = await asyncio.create_subprocess_exec(
@@ -549,7 +586,7 @@ async def stunnel_output(cfg: Config, p_out: asyncio.StreamReader) -> None:
         while True:
             data = await p_out.readline()
             if not data:
-                raise Exception("The stunnel child process ended unexpectedly")
+                raise UnexpectedChildError  # noqa: TRY301
 
             line = data.decode().rstrip("\r\n")
             print(f"[stunnel_output] Read {line!r}")
@@ -559,7 +596,7 @@ async def stunnel_output(cfg: Config, p_out: asyncio.StreamReader) -> None:
     except Exception as err:  # noqa: BLE001
         print(f"[stunnel_output] Complaining to the main thread about {err}")
         await cfg.mainq.put(
-            FatalEvent(etype="stunnel_output", reason=f"Something went wrong: {err}")
+            FatalEvent(etype="stunnel_output", reason=f"Something went wrong: {err}"),
         )
 
 
@@ -681,7 +718,8 @@ def start_connections(  # noqa: PLR0913
         conns.unresolved[idx] = conn
         validate_conns(conns)
         cfg.tasks[f"{prefix}{idx}"] = asyncio.create_task(
-            test_connect(cfg, conn), name=f"test {prefix}{idx}"
+            test_connect(cfg, conn),
+            name=f"test {prefix}{idx}",
         )
 
 
@@ -790,7 +828,8 @@ async def main() -> None:
             print("[main] Launching the stunnel output pipe thread")
             assert proc.stderr is not None  # noqa: S101
             cfg.tasks["output"] = asyncio.create_task(
-                stunnel_output(cfg, proc.stderr), name="collect output"
+                stunnel_output(cfg, proc.stderr),
+                name="collect output",
             )
             print("[main] Waiting for stunnel to start up")
             await expect_event(cfg.mainq, "main", ChildReadyEvent)
